#include <stdlib.h>
#include <math.h>
#include <setjmp.h>
#include <stdio.h>
#include <string.h>
#include "calc.h"

#define ERR(n) {ERROR=n; ERPOS=expression-ERANC-1; strcpy(ERTOK,token); longjmp(jb,1);}

double deg( double x );
double rad( double x );

int   ERROR;
char  ERTOK[TOKLEN + 1];
int   ERPOS;
char* ERANC;


VARIABLE Consts[] =
{
   { "pi",      M_PI },
   { "e",       M_E },

   { 0 }
};


FUNCTION Funcs[] =
{
   { "cos",     1},
   { "sin",     2},
   { "tan",     3},
   { "asin",    4},
   { "acos",    5},
   { "atan",    6},
   { "exp",     10},
   { "log",     11},
   { "log10",   12},
   { "sqrt",    13},
   { "abs",     16},
   { 0 }
};

VARIABLE        Vars[MAXVARS];
unsigned char*  expression;
unsigned char   token[TOKLEN + 1];
int             type;
jmp_buf         jb;


void ClearAllVars(void);
static Parse(void);
static Level1( TYPE* r );
static Level2( TYPE* r );
static Level3( TYPE* r );
static Level4( TYPE* r );
static Level5( TYPE* r );
static Level6( TYPE* r );
Evaluate( char* e, TYPE* result, int* a );

double deg( double x )
{
   return( x * 180.0 / M_PI );
}

double rad( double x )
{
   return( x * M_PI / 180.0 );
}




GetSymbol( char* s, TYPE* v )
{
   char* e;

   e = getenv( s );
   if( !e )
      return( 0 );
   *v = atof( e );
   return( 1 );
}

void ClearAllVars(void)
{
   int i;

   for( i = 0; i < MAXVARS; i++ )
   {
      *Vars[i].name = 0;
       Vars[i].value = 0;
   }
}


ClearVar( char* name )
{
   int i;

   for( i = 0; i < MAXVARS; i++ )
      if( *Vars[i].name && ! strcmp( name, Vars[i].name ) )
      {
         *Vars[i].name = 0;
         Vars[i].value = 0;
         return( 1 );
      }
   return( 0 );
}


GetValue( char* name, TYPE* value )
{
   int i;

   if( *name == '_' )
      return( GetSymbol( name + 1, value ) );

   for( i = 0; i < MAXVARS; i++ )
      if( *Vars[i].name && ! strcmp( name, Vars[i].name ) )
      {
         *value = Vars[i].value;
         return( 1 );
      }

   for( i = 0; *Consts[i].name; i++ )
      if( *Consts[i].name && ! strcmp( name, Consts[i].name ) )
      {
         *value = Consts[i].value;
         return( 1 );
      }
   return( 0 );
}

SetValue( char* name, TYPE* value )
{

   int  i;

  ClearVar( name );
   for( i = 0; i < MAXVARS; i++ )
      if( ! *Vars[i].name )
      {
         strcpy( Vars[i].name, name );
         Vars[i].name[VARLEN] = 0;
         Vars[i].value = *value;
         return( 1 );
      }
   return( 0 );
}

static Parse(void)
{
   char* t;

   type = 0;
   t = token;
   while( iswhite( *expression ) )
      expression++;
   if( isdelim( *expression ) )
   {
      type = DEL;
      *t++ = *expression++;
   }
   else if( isnumer( *expression ) )
   {
      type = NUM;
      while( isnumer( *expression ) )
         *t++ = *expression++;
   }
   else if( isalpha( *expression ) )
   {
      type = VAR;
      while( isalpha( *expression ) )
        *t++ = *expression++;
      token[VARLEN] = 0;
   }
   else if( *expression )
   {
      *t++ = *expression++;
      *t = 0;
      ERR( E_SYNTAX );
   }
   *t = 0;
   while( iswhite( *expression ) )
      expression++;
  return 0;
}

static Level1( TYPE* r )
{
   char t[VARLEN + 1];

   if( type == VAR )
      if( *expression == '=' )
      {
         strcpy( t, token );
         Parse();
         Parse();
         if( !*token )
         {
            ClearVar( t );
            return 1;
         }

         Level2( r );
         if( ! SetValue( t, r ) )
            ERR( E_MAXVARS );
         return 1;
      }
   Level2( r );
   return 0;
}

static Level2( TYPE* r )
{
   TYPE t = 0;
	 char o;

   Level3( r );
   while( (o = *token) == '+' || o == '-' )
   {
      Parse();
      Level3( &t );
      if( o == '+' )
         *r = *r + t;
      else if( o == '-' )
         *r = *r - t;
   }
   return 0;
}

static Level3( TYPE* r )
{
   TYPE t;
   char o;

   Level4( r );
   while( (o = *token) == '*' || o == '/' || o == '%' )
   {
      Parse();
      Level4( &t );
      if( o == '*' )
         *r = *r * t;
      else if( o == '/' )
      {
         if( t == 0 )
            ERR( E_DIVZERO );
         *r = *r / t;
      }
      else if( o == '%' )
      {
         if( t == 0 )
            ERR( E_DIVZERO );
         *r = fmod( *r, t );
      }
   }
   return 0;
}

static Level4( TYPE* r )
{
   TYPE t;

   Level5( r );
   if( *token == '^' )
   {
      Parse();
       Level5( &t );
      *r = pow( *r, t );
   }
   return 0;
}

static Level5( TYPE* r )
{
   char o = 0;

   if( *token == '+' || *token == '-' )
   {
      o = *token;
      Parse();
   }
   Level6( r );
   if( o == '-' )
      *r = -*r;
  return 0;
}

static Level6( TYPE* r )
{
   int  i;
//   int  n;
   TYPE a;


   if( *token == '(' )
   {
      Parse();
      if( *token == ')' )
         ERR( E_NOARG );
      Level1( r );
      if( *token != ')' )
         ERR( E_UNBALAN );
      Parse();
   }
   else
   {
      if( type == NUM )
      {
         *r = (TYPE) atof( token );
         Parse();
      }
      else if( type == VAR )
      {
         if( *expression == '(' )
         {
            for( i = 0; *Funcs[i].name; i++ )
               if( ! strcmp( token, Funcs[i].name ) )
               {
                  Parse();
                  Parse();
                     if( *token == ')' || *token == ',' ) ERR( E_NOARG );
                  Level1( &a );
                  Parse();

				 switch( Funcs[i].idnt)
				 	{
						case 1:
							*r = cos( a*M_PI/180);
						break;

						case 2:
							*r = sin( a*M_PI/180);
						break;

						case 3:
							*r = tan( a*M_PI/180);
						break;

						case 4:
							*r = asin( a)*180/M_PI;
						break;

						case 5:
							*r = acos( a)*180/M_PI;
						break;

						case 6:
							*r = atan( a)*180/M_PI;
						break;

						case 10:
							*r = exp( a);
						break;

						case 11:
							*r = log( a);
						break;

						case 12:
							*r = log10( a);
						break;

						case 13:
							*r = sqrt( a);
						break;

						case 16:
							*r = abs( a);
						break;
					}
                  return 0;
               }

               if( ! *Funcs[i].name )
                  ERR( E_BADFUNC );
            }
            else if( ! GetValue( token, r ) )
               ERR( E_UNKNOWN );
         Parse();
      }
      else
         ERR( E_SYNTAX );
   }
   return 0;
}

Evaluate( char* e, TYPE* result, int* a )
{
   if( setjmp( jb ) )
      return( ERROR );
   expression = e;
   ERANC = e;
   strlwr( expression );
   *result = 0;
   Parse();
   if( ! *token )
   ERR( E_EMPTY );
   *a = Level1( result );
   return( E_OK );
}


//main( int argc, char* argv[] )
//{
//	char data[1024];
//	double res;
//	int ercode;
//printf("Enter data -> ");
//gets(data);
//Evaluate(data,&res,&ercode);
//printf("result: %f",res);
//   return 0;
//}
